/*------------------------ BEGIN small_blas.c BEGIN -----------------------*/
/* @file small_blas.c                                                      */
/* @author jlinford                                                        */
/* @date 2012-06-22 15:41:00.217859                                        */
/* @brief Basic linear algebra subprogram definitions                      */
/*                                                                         */
/* A reduced set of BLAS routines optimized for Kppa-generated solvers     */
/*                                                                         */
/* This file was generated by Kppa: http://www.paratools.com/Kppa          */
/*-------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include "small_parameters.h"
#include "small_blas.h"



/*--------------------------------- WCOPY ---------------------------------*/
/* Copies vector x to vector y: y <= x                                     */
/* Like the BLAS {S,D}COPY(N,X,1,Y,1)                                      */
/*                                                                         */
/* @param[in]     n Vector length                                          */
/* @param[in]     x Vector x                                               */
/* @param[out]    y Vector y                                               */
/*-------------------------------------------------------------------------*/
void WCOPY(unsigned long const n, float x[], float y[])
{
    memcpy(y, x, n*sizeof(float));
} /* END WCOPY */


/*--------------------------------- WSCAL ---------------------------------*/
/* Constant times a vector: x <= alpha*x                                   */
/* Like the BLAS {S,D}SCAL(N,alpha,X,1)                                    */
/*                                                                         */
/* @param[in]     n     Vector length                                      */
/* @param[in]     alpha Scalar                                             */
/* @param[in,out] x     Vector x                                           */
/*-------------------------------------------------------------------------*/
void WSCAL(unsigned long const n, float const alpha, float x[])
{
    int blk;
    int rem;

    if(alpha == 0.0) {
        memset(x, 0, n*sizeof(float));
    } else if(alpha == 1.0) {
        return;
    } else if(alpha == -1.0) {
        blk = n >> 3;
        rem = n & 7;
        while(blk) {
            x[0] = -x[0];
            x[1] = -x[1];
            x[2] = -x[2];
            x[3] = -x[3];
            x[4] = -x[4];
            x[5] = -x[5];
            x[6] = -x[6];
            x[7] = -x[7];
            --blk;
            x += 8;
        }
        switch(rem) {
        case 7: x[6] = -x[6];
        case 6: x[5] = -x[5];
        case 5: x[4] = -x[4];
        case 4: x[3] = -x[3];
        case 3: x[2] = -x[2];
        case 2: x[1] = -x[1];
        case 1: x[0] = -x[0];
        }
    } else {
        blk = n >> 3;
        rem = n & 7;
        while(blk) {
            x[0] *= alpha;
            x[1] *= alpha;
            x[2] *= alpha;
            x[3] *= alpha;
            x[4] *= alpha;
            x[5] *= alpha;
            x[6] *= alpha;
            x[7] *= alpha;
            --blk;
            x += 8;
        }
        switch(rem) {
        case 7: x[6] *= alpha;
        case 6: x[5] *= alpha;
        case 5: x[4] *= alpha;
        case 4: x[3] *= alpha;
        case 3: x[2] *= alpha;
        case 2: x[1] *= alpha;
        case 1: x[0] *= alpha;
        }
    }
} /* END WSCAL */


/*--------------------------------- WAXPY ---------------------------------*/
/* Constant times a vector plus a vector: y <= y + alpha*x                 */
/* Like the BLAS {S,D}AXPY(N,alpha,X,1,Y,1)                                */
/*                                                                         */
/* @param[in]     n     Vector length                                      */
/* @param[in]     alpha Scalar                                             */
/* @param[in]     x     Vector x                                           */
/* @param[in,out] y     Vector y                                           */
/*-------------------------------------------------------------------------*/
void WAXPY(unsigned long const n, float const alpha, float x[], float y[])
{
    int blk;
    int rem;

    if(alpha == ZERO) {
        return;
    } else if(alpha == ONE) {
        blk = n >> 3;
        rem = n & 7;
        while(blk) {
            y[0] += x[0];
            y[1] += x[1];
            y[2] += x[2];
            y[3] += x[3];
            y[4] += x[4];
            y[5] += x[5];
            y[6] += x[6];
            y[7] += x[7];
            --blk;
            x += 8;
            y += 8;
        }
        switch(rem) {
        case 7: y[6] += x[6];
        case 6: y[5] += x[5];
        case 5: y[4] += x[4];
        case 4: y[3] += x[3];
        case 3: y[2] += x[2];
        case 2: y[1] += x[1];
        case 1: y[0] += x[0];
        }
    } else if(alpha == -ONE) {
        blk = n >> 3;
        rem = n & 7;
        while(blk) {
            y[0] -= x[0];
            y[1] -= x[1];
            y[2] -= x[2];
            y[3] -= x[3];
            y[4] -= x[4];
            y[5] -= x[5];
            y[6] -= x[6];
            y[7] -= x[7];
            --blk;
            x += 8;
            y += 8;
        }
        switch(rem) {
        case 7: y[6] -= x[6];
        case 6: y[5] -= x[5];
        case 5: y[4] -= x[4];
        case 4: y[3] -= x[3];
        case 3: y[2] -= x[2];
        case 2: y[1] -= x[1];
        case 1: y[0] -= x[0];
        }
    } else {
        blk = n >> 3;
        rem = n & 7;
        while(blk) {
            y[0] += alpha * x[0];
            y[1] += alpha * x[1];
            y[2] += alpha * x[2];
            y[3] += alpha * x[3];
            y[4] += alpha * x[4];
            y[5] += alpha * x[5];
            y[6] += alpha * x[6];
            y[7] += alpha * x[7];
            --blk;
            x += 8;
            y += 8;
        }
        switch(rem) {
        case 7: y[6] += alpha * x[6];
        case 6: y[5] += alpha * x[5];
        case 5: y[4] += alpha * x[4];
        case 4: y[3] += alpha * x[3];
        case 3: y[2] += alpha * x[2];
        case 2: y[1] += alpha * x[1];
        case 1: y[0] += alpha * x[0];
        }
    }
} /* END WAXPY */


/*--------------------------------- WYMXDA --------------------------------*/
/* Difference of two vectors divided by a constant: z <= (y - x) / alpha   */
/*                                                                         */
/* @param[in]     n     Vector length                                      */
/* @param[in]     x     Vector x                                           */
/* @param[in]     y     Vector y                                           */
/* @param[in]     alpha Scalar                                             */
/* @param[out]    z     Vector z                                           */
/*-------------------------------------------------------------------------*/
void WYMXDA(unsigned long const n, float x[], float y[], float const alpha,
        float z[])
{
    int blk = n >> 3;
    int rem = n & 7;

    if(alpha == 1.0) {
        while(blk) {
            z[0] = y[0] - x[0];
            z[1] = y[1] - x[1];
            z[2] = y[2] - x[2];
            z[3] = y[3] - x[3];
            z[4] = y[4] - x[4];
            z[5] = y[5] - x[5];
            z[6] = y[6] - x[6];
            z[7] = y[7] - x[7];
            --blk;
            x += 8;
            y += 8;
            z += 8;
        }
        switch(rem) {
        case 7: z[6] = y[6] - x[6];
        case 6: z[5] = y[5] - x[5];
        case 5: z[4] = y[4] - x[4];
        case 4: z[3] = y[3] - x[3];
        case 3: z[2] = y[2] - x[2];
        case 2: z[1] = y[1] - x[1];
        case 1: z[0] = y[0] - x[0];
        }
    } else {
        while(blk) {
            z[0] = (y[0] - x[0])/alpha;
            z[1] = (y[1] - x[1])/alpha;
            z[2] = (y[2] - x[2])/alpha;
            z[3] = (y[3] - x[3])/alpha;
            z[4] = (y[4] - x[4])/alpha;
            z[5] = (y[5] - x[5])/alpha;
            z[6] = (y[6] - x[6])/alpha;
            z[7] = (y[7] - x[7])/alpha;
            --blk;
            x += 8;
            y += 8;
            z += 8;
        }
        switch(rem) {
        case 7: z[6] = (y[6] - x[6])/alpha;
        case 6: z[5] = (y[5] - x[5])/alpha;
        case 5: z[4] = (y[4] - x[4])/alpha;
        case 4: z[3] = (y[3] - x[3])/alpha;
        case 3: z[2] = (y[2] - x[2])/alpha;
        case 2: z[1] = (y[1] - x[1])/alpha;
        case 1: z[0] = (y[0] - x[0])/alpha;
        }
    }
} /* END WYMXDA */


/*-------------------------- END small_blas.h END -------------------------*/
