/*----------------------- BEGIN small_rates.c BEGIN -----------------------*/
/* @file small_rates.c                                                     */
/* @author jlinford                                                        */
/* @date 2012-06-22 15:41:00.251845                                        */
/* @brief Reaction rate calculation and utility functions                  */
/*                                                                         */
/* Reaction rate calculation and utility functions                         */
/*                                                                         */
/* This file was generated by Kppa: http://www.paratools.com/Kppa          */
/*-------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include "small_parameters.h"
#include "small_rates.h"


/*---------------------------------- ARR ----------------------------------*/
/* @param[in]     a0   None                                                */
/* @param[in]     b0   None                                                */
/* @param[in]     c0   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
float ARR(float const a0, float const b0, float const c0, float const temp)
{
    return a0 * pow(temp/(float)300.0, c0) * exp(-b0/temp);
} /* END ARR */


/*---------------------------------- ARR2 ---------------------------------*/
/* @param[in]     a0   None                                                */
/* @param[in]     b0   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
float ARR2(float const a0, float const b0, float const temp)
{
    return a0 * exp(b0/temp);
} /* END ARR2 */


/*---------------------------------- EP2 ----------------------------------*/
/* @param[in]     a0   None                                                */
/* @param[in]     c0   None                                                */
/* @param[in]     a2   None                                                */
/* @param[in]     c2   None                                                */
/* @param[in]     a3   None                                                */
/* @param[in]     c3   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
float EP2(float const a0, float const c0, float const a2, float const c2,
        float const a3, float const c3, float const temp)
{
    float k0 = a0 * exp(-c0/temp);
    float k2 = a2 * exp(-c2/temp);
    float k3 = a3 * exp(-c3/temp) * (float)1.0e6*CFACTOR;
    return k0 + k3/((float)1.0 + k3/k2);
} /* END EP2 */


/*---------------------------------- EP3 ----------------------------------*/
/* @param[in]     a1   None                                                */
/* @param[in]     c1   None                                                */
/* @param[in]     a2   None                                                */
/* @param[in]     c2   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
float EP3(float const a1, float const c1, float const a2, float const c2,
        float const temp)
{
    float k1 = a1 * exp(-c1/temp);
    float k2 = a2 * exp(-c2/temp) * (float)1.0e6*CFACTOR;
    return k1 + k2;
} /* END EP3 */


/*---------------------------------- FALL ---------------------------------*/
/* @param[in]     a0   None                                                */
/* @param[in]     b0   None                                                */
/* @param[in]     c0   None                                                */
/* @param[in]     a1   None                                                */
/* @param[in]     b1   None                                                */
/* @param[in]     c1   None                                                */
/* @param[in]     cf   None                                                */
/* @param[in]     temp Temperature                                         */
/*-------------------------------------------------------------------------*/
float FALL(float const a0, float const b0, float const c0, float const a1,
        float const b1, float const c1, float const cf, float const temp)
{
    float k0 = a0 * pow(temp/(float)300.0, c0) * exp(-b0/temp) * (float)1.0e6*CFACTOR; 
    float k1 = k0 / (a1 * pow(temp/(float)300.0, c1) * exp(-b1/temp));
    return (k0/((float)1.0+k1)) * pow(cf, (float)1.0/((float)1.0+pow(log10(k1),(float)2.0)));
} /* END FALL */


/*-------------------------------- Sunlight -------------------------------*/
/* Calculates sunlight intensity in the range [0,1] as a function of time. */
/* Modify this routine to get the correct sunlight values for your model.  */
/*                                                                         */
/* @param[in]     time Integration time                                    */
/* @param[in]     idx  Current grid cell index                             */
/*-------------------------------------------------------------------------*/
float Sunlight(float const time, unsigned long const idx)
{
    int daysec = 24 * 3600;     /* Seconds per day */
    float sunrise = 5.5*3600;   /* 5:30 local time */
    float sunset = 19.5*3600;   /* 7:30 local time */

    float daily = time - ((int)time / daysec) * daysec;
    float tmp;

    /* Estimate sunlight intensity in the range [0,1] */
    if ((daily >= sunrise) && (daily <= sunset)) {
        tmp = (2.0 * daily - sunrise - sunset) / (sunset - sunrise);
        tmp = (tmp > 0) ? tmp * tmp : -tmp * tmp;
        tmp = 0.5 * (1.0 + cos(M_PI * tmp));
    } else {
        tmp = 0.0;
    }
    return tmp;
} /* END Sunlight */


/*------------------------------ Temperature ------------------------------*/
/* Calculates temperature (kelvin) as a function of time.                  */
/* Modify this routine to get the correct temperature values for your model. */
/*                                                                         */
/* @param[in]     time Integration time                                    */
/* @param[in]     idx  Current grid cell index                             */
/*-------------------------------------------------------------------------*/
float Temperature(float const time, unsigned long const idx)
{
    float mintemp = 280;    /* 280 Kelvin ~= 44 Fahrenheit */
    float maxtemp = 300;    /* 300 Kelvin ~= 80 Fahrenheit */
    float tmp;

    /* Estimate temperature cycling from mintemp to maxtemp */
    tmp = sin(time*M_PI/(24*3600));
    if (tmp < 0) {
        tmp = mintemp - tmp * (maxtemp-mintemp);
    } else {
        tmp = mintemp + tmp * (maxtemp-mintemp);
    }
    return tmp;
} /* END Temperature */


/*--------------------------------- Rates ---------------------------------*/
/* Calculates reaction rate coefficients                                   */
/*                                                                         */
/* @param[in]     time Integration time                                    */
/* @param[in]     idx  Current grid cell index                             */
/* @param[out]    rct  Reaction rates                                      */
/*-------------------------------------------------------------------------*/
void Rates(float const time, unsigned long const idx, float rct[10])
{
    /* Sunlight intensity: 0 to 1 inclusive (uppercase for KPP compatibility) */
    float SUN;
    /* Temperature in kelvin  (uppercase for KPP compatibility) */
    float TEMP;

    SUN = Sunlight(time, idx);
    TEMP = Temperature(time, idx);

    rct[0] = (2.643E-10) * SUN*SUN*SUN;
    rct[1] = 8.018e-17;
    rct[2] = (6.120E-04) * SUN;
    rct[3] = 1.576e-15;
    rct[4] = (1.070E-03) * SUN*SUN;
    rct[5] = 7.11e-11;
    rct[6] = 1.2e-10;
    rct[7] = 6.062e-15;
    rct[8] = 1.069e-11;
    rct[9] = (1.289E-02) * SUN;
} /* END Rates */



/*------------------------- END small_rates.h END -------------------------*/
